<?php

namespace PdToolKit\Utils;

use Illuminate\Support\Arr;

/**
 * Class Decimal
 * @package PdToolKit\Utils
 * @method Decimal add(int|float|string|Decimal $v);
 * @method Decimal sub(int|float|string|Decimal $v);
 * @method Decimal mul(int|float|string|Decimal $v);
 * @method Decimal div(int|float|string|Decimal $v);
 */
class Decimal
{
    protected $source;
    protected $dp;

    // Rounds away from zero.
    const ROUND_UP = 0;
    // Rounds towards zero.
    const ROUND_DOWN = 1;

    function __construct($v, int $dp = 8)
    {
        if (!$this->isNumeric($v)) {
            throw new \Exception("value type must be int|float|string|Decimal.");
        }

        $this->dp = $dp;
        $val = $v instanceof Decimal ? $v->getSource() : $v;
        $this->source = bcadd($val, "0", $dp);
    }

    private function isNumeric($v)
    {
        if ($v instanceof Decimal) return true;
        return is_numeric($v);
    }

    function getSource()
    {
        return $this->source;
    }

    function __call($name, $arguments)
    {
        $methods = ['add', 'sub', 'mul', 'div'];
        if (in_array($name, $methods)) {
            $method = "bc$name";
            $v = new Decimal($arguments[0]);
            return new Decimal($method($this->source, $v->getSource()));
        }

        throw new \Exception("method $name not found.");
    }

    function __toString(): string
    {
        return $this->source;
    }

    function equals($v): bool
    {
        return (new Decimal($v))->getSource() == new Decimal($this->source);
    }

    function toDecimalPlaces(int $dp, int $rm = self::ROUND_UP): Decimal
    {
        return new Decimal($this->toFixed($dp, $rm), $dp);
    }

    function toFixed(int $dp, int $rm = self::ROUND_UP): string
    {
        if ($rm == self::ROUND_UP) {
            return number_format($this->source, $dp, ".", "");
        }

        if ($rm == self::ROUND_DOWN) {
            return bcadd($this->source, 0, $dp);
        }

        throw new \Exception("$rm must be " . self::ROUND_UP . ' or ' . self::ROUND_DOWN);
    }

    function toInt(): int
    {
        return (int) $this->source;
    }

    function toFloat(): float
    {
        return (float) $this->source;
    }
}
